# Modified from: https://github.com/pliang279/LG-FedAvg/blob/master/main_fed.py
# credit goes to: Paul Pu Liang

#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Python version: 3.6

# This program implements FedRep under the specification --alg fedrep, as well as Fed-Per (--alg fedper), LG-FedAvg (--alg lg), 
# FedAvg (--alg fedavg) and FedProx (--alg prox)

import copy
import itertools
import numpy as np
import pandas as pd
import torch
from torch import nn

from untils.options import args_parser
#from utils_fedrep.train_utils import get_data, get_model, read_data
from untils.Update_pFedSeg import LocalUpdatepFedCS
from untils.utils import average_weights, exp_details, get_dataset
# from models.test import test_img_local_all
import model.model as ml
import model.encoder as encoder
import time

if __name__ == '__main__':
    # parse args
    args = args_parser()
    lens = np.ones(args.num_users)
    args.device = torch.device('cuda:{}'.format(args.gpu) if torch.cuda.is_available() and args.gpu != -1 else 'cpu')
    print(args.device)
    # ##### fix random seeds for reproducibility ########
    SEED = args.seed
    torch.manual_seed(SEED)
    torch.backends.cudnn.deterministic = False
    torch.backends.cudnn.benchmark = False
    np.random.seed(SEED)
    #####################################################


    if args.task == "PMR":
        if args.model == 'UNet':
            global_model = ml.UNet(1,2)
            Encoder = encoder.UNet(1,2)
            #print("ooooooo")
        else:
            exit('Error: unrecognized model')
    else:
        if args.model == 'UNet':
            global_model = ml.UNet(3,2)
            Encoder = encoder.UNet(3,2)
            #print("ooooooo")
        else:
            exit('Error: unrecognized model')
    print(args.alg)

    # build model
    global_model.train()

    total_num_layers = len(global_model.state_dict().keys())
    #print(global_model.state_dict().keys())
    net_keys = [*global_model.state_dict().keys()]

    # specify the representation parameters (in w_glob_keys) and head parameters (all others)
    if args.alg == 'pFedCSeg':
       w_glob_keys = net_keys[:-58]
       w_glob_keys_body = net_keys[:-2]

    print(total_num_layers)
    print(len(w_glob_keys)) ###少两层
    #print(net_keys)
    #####参数量计算
    num_param_glob = 0
    num_param_local = 0
    for key in global_model.state_dict().keys():
        num_param_local += global_model.state_dict()[key].numel()
        #print(num_param_local)
        if key in w_glob_keys:
            num_param_glob += global_model.state_dict()[key].numel()
    percentage_param = 100 * float(num_param_glob) / num_param_local
    print('# Params: {} (local), {} (global); Percentage {:.2f} ({}/{})'.format(
        num_param_local, num_param_glob, percentage_param, num_param_glob, num_param_local))
    print("learning rate, batch size: {}, {}".format(args.lr, args.local_bs))
    #####
    
    net_keys = [*Encoder.state_dict().keys()]
    # print(Encoder.up4)
    print(len(net_keys))
    i = 0
    for key in Encoder.state_dict().keys():
        if key in w_glob_keys:
           i=i+1
    print(i)         
    
    net_keys = [*Encoder.state_dict().keys()]

    # generate list of local models for each user
    net_local_list = []
    w_locals = {}
    for user in range(args.num_users):
        w_local_dict = {}
        for key in global_model.state_dict().keys():
            w_local_dict[key] =global_model.state_dict()[key]
        #print("aaaaaaaa",len(w_local_dict))
        w_locals[user] = w_local_dict



    w_locals_encoder = {}
    for user in range(args.num_users):
        w_local_dict = {}
        for key in Encoder.state_dict().keys():
            w_local_dict[key] =Encoder.state_dict()[key]
        #print("aaaaaaaa",len(w_local_dict))
        w_locals_encoder[user] = w_local_dict


  


    # training
    indd = None     
    loss_train = []
    accs = []
    times = []
    accs10 = 0
    accs10_glob = 0
    start = time.time()
    train_loss, train_accuracy = [], []
    encoder_train_epoch = 10
    for iter in range(args.epochs+1):
        w_glob = {}
        loss_locals = []
        if args.task == "RIF":
           idxs_users = [0,1,2,3,4,5,6]#np.random.choice(range(args.num_users), m, replace=False)
        if args.task == "EndoPolyp":
           idxs_users = [0,1,2,3]#np.random.choice(range(args.num_users), m, replace=False)        
        if args.task == "PMR":
           idxs_users = [0,1,2,3,4,5]#np.random.choice(range(args.num_users), m, replace=False)        
        #print(idxs_users)
        w_keys_epoch = w_glob_keys
        times_in = []
        total_len=0
        for ind, idx in enumerate(idxs_users):
            start_in = time.time()
            if args.task == "EndoPolyp":
                if (iter%2) == 0:
                   args.lr = 0.00001
                else:
                   args.lr = 0.0001
            if args.task == "RIF":
                if idx ==5:
                   args.lr =0.00001
                else:
                   args.lr =0.0001
            if args.task == "PMR":
                if (iter%2) == 0:
                   args.lr = 0.00001
                else:
                   args.lr = 0.0001
            train_dataset, test_dataset = get_dataset(idx,args.task)
            #local = LocalUpdatepFedSeg(args, train_dataset, test_dataset, idx)
            local = LocalUpdatepFedCS(args, train_dataset, test_dataset, idx)
            total_len += lens[idx]
            if (iter%2) == 0:
               #args.local_ep = 5
               net_local = copy.deepcopy(Encoder)
               w_local = net_local.state_dict()
               seg_model = copy.deepcopy(global_model)
               seg_model_glob = seg_model.state_dict()
               for k in w_locals_encoder[idx].keys():####w_locals所有参数
                    if k in w_glob_keys: 
                        w_local[k] = seg_model_glob[k]
                        #print("bbbbbb")
                    else:
                        w_local[k] = w_locals_encoder[idx][k]
                    # else:
                        # w_local[k] = encoder_glob[k]
                        #print("aaaaaaaaa")
               net_local.load_state_dict(w_local)
               
               w_local, loss = local.train_body(net=net_local.to(args.device), w_glob_keys=w_glob_keys, global_round=iter, lr=args.lr)
               
               if len(w_glob) == 0:
                    w_glob = copy.deepcopy(w_local)
                    for k,key in enumerate(Encoder.state_dict().keys()):
                        w_glob[key] = w_glob[key]*lens[idx]
                        w_locals_encoder[idx][key] = w_local[key]
               else:
                    for k,key in enumerate(Encoder.state_dict().keys()):
                        if key in w_glob_keys:   ####only encoder
                            w_glob[key] += w_local[key]*lens[idx]
                        else:
                            w_glob[key] += w_local[key]*lens[idx]
                        w_locals_encoder[idx][key] = w_local[key]
            else:
               #args.local_ep = 5
               net_local = copy.deepcopy(global_model)
               w_local = net_local.state_dict()
               encoder_model = copy.deepcopy(Encoder)
               encoder_glob = encoder_model.state_dict()
               for k in w_locals[idx].keys():####w_locals所有参数
                    if k not in w_glob_keys: 
                        w_local[k] = w_locals[idx][k]
                        #print("bbbbbb")
                    else:
                        w_local[k] = encoder_glob[k]
                        #print("aaaaaaaaa")
               net_local.load_state_dict(w_local)
               w_local, loss = local.train_local(net=net_local.to(args.device), w_glob_keys=w_glob_keys, global_round=iter, lr=args.lr)
               net_local.load_state_dict(w_local)
               #local.inference(net_local.to(args.device), None, None, idx, args.task)
               if len(w_glob) == 0:
                    w_glob = copy.deepcopy(w_local)
                    for k,key in enumerate(global_model.state_dict().keys()):
                        w_glob[key] = w_glob[key]*lens[idx]
                        w_locals[idx][key] = w_local[key]
               else:
                    for k,key in enumerate(global_model.state_dict().keys()):
                        if key in w_glob_keys:   ####only encoder
                            w_glob[key] += w_local[key]*lens[idx]
                        else:
                            w_glob[key] += w_local[key]*lens[idx]
                        w_locals[idx][key] = w_local[key]

        # get weighted average for global weights
        if (iter%2) == 0:
            for k in Encoder.state_dict().keys():
                w_glob[k] = torch.div(w_glob[k], total_len)
            Encoder.load_state_dict(w_glob)
        else:
            for k in global_model.state_dict().keys():
                w_glob[k] = torch.div(w_glob[k], total_len)
            global_model.load_state_dict(w_glob)
            #if ((iter+1) % 1) == 0:
        ###测试每个测试集的结果
            list_acc, list_loss = [], []
            for ind, idx in enumerate(idxs_users):
                train_dataset, test_dataset = get_dataset(idx,args.task)
                local_model = LocalUpdatepFedCS(args, train_dataset, test_dataset, idx)
                local_model.inference(global_model, w_locals, None, idx, args.task, iter)
            # if ((iter+1) % 1) == 0:
            # ####测试每个测试集的结果
                # dice_list = []
                # loss_list = []
                # list_acc, list_loss = [], []
                # for ind, idx in enumerate(idxs_users):
                    # train_dataset, test_dataset = get_dataset(idx,args.task)
                    # local_model = LocalUpdatepFedCS(args, train_dataset, test_dataset, idx)
                    # dice,loss = local_model.inference_average(global_model, w_locals, None, idx)
                    # dice_list.append(dice.item())
                    # loss_list.append(loss.item())
                # print(iter,dice_list)
                
"""
        # for k in global_model.state_dict().keys():
            # w_glob[k] = torch.div(w_glob[k], total_len)

        w_local = global_model.state_dict()
        for k in w_glob.keys():
            w_local[k] = w_glob[k]
        # if args.epochs != iter:
            # global_model.load_state_dict(w_glob)
        
"""
