import torch
from torch import nn
from torch.utils.data import DataLoader, Dataset
import math
import numpy as np
import time
import copy
import random
from torchvision import datasets, transforms
from PIL import Image
from untils.dataread import data_read,data_read_EndoPolyp,data_read_PMR
from untils.dice_score import dice_coeff,dice_loss,multiclass_dice_coeff
import torch.nn.functional as F
from torchvision import transforms as T


class DatasetSplit(Dataset):
    """An abstract Dataset class wrapped around Pytorch Dataset class.
    """

    def __init__(self, path, task):
        self.dataset = path[:]
        self.task = task
    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, item): 
        if self.task == "RIF":
            image, label = data_read(self.dataset[item])
        if self.task == "EndoPolyp":
            image, label = data_read_EndoPolyp(self.dataset[item])
        if self.task == "PMR":
            image, label = data_read_PMR(self.dataset[item])
        #print(image.dtype)
        #print(torch.tensor(image))
        return image, label

class Datasettest(Dataset):
    """An abstract Dataset class wrapped around Pytorch Dataset class.
    """

    def __init__(self, path):
        self.dataset = path[:]

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, item):
        image, label = data_read(self.dataset[item])
        #print(image.dtype)
        #print(torch.tensor(image))
        return torch.tensor(image), torch.tensor(label)
        
        
class LocalUpdatepFedSeg(object):
    def __init__(self, args, dataset_train, dataset_test, idx):
        self.args = args
        self.loss_func = nn.CrossEntropyLoss()
        self.trainloader, self.testloader = self.train_test(dataset_train,dataset_test)
        self.idx=idx
        
    def train_test(self, dataset_train, dataset_test):
        """
        Returns train, validation and test dataloaders for a given dataset
        and user indexes.
        """
        trainloader = DataLoader(DatasetSplit(dataset_train,self.args.task),batch_size=self.args.local_bs, shuffle=True)
        testloader = DataLoader(DatasetSplit(dataset_test,self.args.task),batch_size=1, shuffle=True)
        return trainloader, testloader



    def train_local(self, net, w_glob_keys, global_round,lr=0.00001):
        net.train()
        
        if self.args.optimizer == 'sgd':
            optimizer = torch.optim.SGD(net.parameters(), lr=self.args.lr,
                                        momentum=0.5)
        elif self.args.optimizer == 'adam':
            optimizer = torch.optim.Adam(net.parameters(), lr=self.args.lr,
                                         weight_decay=1e-4)

        local_eps = self.args.local_ep
        head_eps = local_eps-self.args.local_rep_ep
        epoch_loss = []
        num_updates = 0
        for iter in range(20):
            # for name, param in net.named_parameters():
                # if name in w_glob_keys:
                    # param.requires_grad = True
                # else:
                    # param.requires_grad = False

            batch_loss = []
            for batch_idx, (images, labels) in enumerate(self.trainloader):
                images, labels = images.to(self.args.device), labels.to(self.args.device)
                net.zero_grad()
                out = net(images)
                loss = dice_loss(
                                    F.softmax(out, dim=1).float(),
                                    F.one_hot(labels,2).permute(0, 3, 1, 2).float(),
                                    multiclass=True)
                loss = 0.5*loss + 0.5*self.loss_func(out, labels)
                loss.backward()
                optimizer.step()
                num_updates += 1
                batch_loss.append(loss.item())
                if batch_idx % 2 == 0:
                    print('\r','| Global Round : {} | Local Epoch : {} | [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                        global_round, iter, batch_idx * len(images),
                        len(self.trainloader.dataset),
                        100. * batch_idx / len(self.trainloader), loss.item()),end='')
            epoch_loss.append(sum(batch_loss)/len(batch_loss))
        print("         ")
        return net.state_dict(), sum(epoch_loss) / len(epoch_loss)



    def train_head(self, net, w_glob_keys, global_round,lr=0.00001):
        net.train()
        
        if self.args.optimizer == 'sgd':
            optimizer = torch.optim.SGD(net.parameters(), lr=self.args.lr,
                                        momentum=0.5)
        elif self.args.optimizer == 'adam':
            optimizer = torch.optim.Adam(net.parameters(), lr=self.args.lr,
                                         weight_decay=1e-4)

        local_eps = self.args.local_ep
        epoch_loss = []
        num_updates = 0
        
        for name, param in net.named_parameters():
            if name in w_glob_keys:
                param.requires_grad = False
            else:
                param.requires_grad = True
        for iter in range(local_eps):
            batch_loss = []
            for batch_idx, (images, labels) in enumerate(self.trainloader):
                images, labels = images.to(self.args.device), labels.to(self.args.device)
                net.zero_grad()
                out = net(images)
                loss = dice_loss(
                                    F.softmax(out, dim=1).float(),
                                    F.one_hot(labels,2).permute(0, 3, 1, 2).float(),
                                    multiclass=True)
                loss = 0.5*loss + 0.5*self.loss_func(out, labels)
                loss.backward()
                optimizer.step()
                num_updates += 1
                batch_loss.append(loss.item())
                if batch_idx % 2 == 0:
                    print('\r','| Global Round : {} | Local Epoch : {} | [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                        global_round, iter, batch_idx * len(images),
                        len(self.trainloader.dataset),
                        100. * batch_idx / len(self.trainloader), loss.item()),end='')
            epoch_loss.append(sum(batch_loss)/len(batch_loss))
        print("         ")
        return net.state_dict(), sum(epoch_loss) / len(epoch_loss)







    def train_body(self, net, w_glob_keys, global_round,lr=0.00001):
        net.train()
        
        if self.args.optimizer == 'sgd':
            optimizer = torch.optim.SGD(net.parameters(), lr=self.args.lr,
                                        momentum=0.5)
        elif self.args.optimizer == 'adam':
            optimizer = torch.optim.Adam(net.parameters(), lr=self.args.lr,
                                         weight_decay=1e-4)

        local_eps = self.args.local_ep
        epoch_loss = []
        num_updates = 0
        
        for name, param in net.named_parameters():
            if name in w_glob_keys:
                param.requires_grad = True
            else:
                param.requires_grad = False
        for iter in range(local_eps):
            batch_loss = []
            for batch_idx, (images, labels) in enumerate(self.trainloader):
                images, labels = images.to(self.args.device), labels.to(self.args.device)
                net.zero_grad()
                out = net(images)
                loss = dice_loss(
                                    F.softmax(out, dim=1).float(),
                                    F.one_hot(labels,2).permute(0, 3, 1, 2).float(),
                                    multiclass=True)
                loss = 0.5*loss + 0.5*self.loss_func(out, labels)
                loss.backward()
                optimizer.step()
                num_updates += 1
                batch_loss.append(loss.item())
                if batch_idx % 2 == 0:
                    print('\r','| Global Round : {} | Local Epoch : {} | [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                        global_round, iter, batch_idx * len(images),
                        len(self.trainloader.dataset),
                        100. * batch_idx / len(self.trainloader), loss.item()),end='')
            epoch_loss.append(sum(batch_loss)/len(batch_loss))
        print("         ")
        return net.state_dict(), sum(epoch_loss) / len(epoch_loss)


    def inference(self, net, w_locals, w_glob_keys, idx):
        """ Returns the inference accuracy and loss.
        """
        model = copy.deepcopy(net)
        if w_locals is not None:
            w_local = model.state_dict()
            for k in w_locals[idx].keys():
                if w_glob_keys is not None and k not in w_glob_keys:
                    w_local[k] = w_locals[idx][k]
                elif w_glob_keys is None:
                    w_local[k] = w_locals[idx][k]
            model.load_state_dict(w_local)
        model = model.to(self.args.device)
        model.eval()
        loss, total, correct = 0.0, 0.0, 0.0

        with torch.no_grad():
            dice_list = []
            for i,dataset in enumerate(self.testloader):
                
                batchimages,batchLabel = dataset
                inputs = batchimages.to(self.args.device)  #######Train
                target = batchLabel.to(self.args.device)
                mask_true = target.long()
                mask_pred = model(inputs)             #得到前向传播的结果
                #print(out.size())
                mask_true = F.one_hot(mask_true, 2).permute(0, 3, 1, 2).float()
                mask_pred = F.one_hot(mask_pred.argmax(dim=1), 2).permute(0, 3, 1, 2).float()
                dice = multiclass_dice_coeff(mask_pred[:, 1:], mask_true[:, 1:], reduce_batch_first=False)
                #print(mask_true.size(),mask_pred.size())
                #print(dice, dice.item())
                dice_list.append(dice.item())
                #print(dice)
                if self.idx == 0:
                    print('\r','BinRushed dice:{:.4f}'.format(np.mean(dice_list)),end='')
                if self.idx == 1:
                    print('\r','Magrabia dice:{:.4f}'.format(np.mean(dice_list)),end='')
                if self.idx == 2:
                    print('\r','IDRID_Train dice:{:.4f}'.format(np.mean(dice_list)),end='')
                if self.idx == 3:
                    print('\r','RIM-ONE dice:{:.4f}'.format(np.mean(dice_list)),end='')
                if self.idx == 4:
                    print('\r','MESSIDOR dice:{:.4f}'.format(np.mean(dice_list)),end='')
                if self.idx == 5:
                    print('\r','REFUGE_Train dice:{:.4f}'.format(np.mean(dice_list)),end='')
                if self.idx == 6:
                    print('\r','Drishti_GS1_Train dice:{:.4f}'.format(np.mean(dice_list)),end='')
            print("         ")
        
        
        
        
        

def default(val, def_val):
    return def_val if val is None else val

def set_requires_grad(model, val):
    for p in model.parameters():
        p.requires_grad = val


def get_target_encoder(online_encoder):
    target_encoder = copy.deepcopy(online_encoder)
    set_requires_grad(target_encoder, False)
    return target_encoder
# loss fn

def loss_fn(x, y):
    x = F.normalize(x, dim=-1, p=2)
    y = F.normalize(y, dim=-1, p=2)
    return 2 - 2 * (x * y).sum(dim=-1)
    
# augmentation utils

class RandomApply(nn.Module):
    def __init__(self, fn, p):
        super().__init__()
        self.fn = fn
        self.p = p
    def forward(self, x):
        if random.random() > self.p:
            return x
        return self.fn(x)
        

class EMA():
    def __init__(self, beta):
        super().__init__()
        self.beta = beta

    def update_average(self, old, new):
        if old is None:
            return new
        return old * self.beta + (1 - self.beta) * new

def update_moving_average(ema_updater, ma_model, current_model):
    for current_params, ma_params in zip(current_model.parameters(), ma_model.parameters()):
        old_weight, up_weight = ma_params.data, current_params.data
        ma_params.data = ema_updater.update_average(old_weight, up_weight)


class LocalUpdatepFedCS(object):
    def __init__(self, args, dataset_train, dataset_test, idx):
        self.args = args
        self.loss_func = nn.CrossEntropyLoss()
        self.trainloader, self.testloader = self.train_test(dataset_train,dataset_test)
        self.idx=idx
        ####数据增强
        if self.args.task == "PMR":
            self.apply_transform =  torch.nn.Sequential(
            #T.Normalize(mean = [0.31955430042667393], std = [0.25145716010016883]),
            )
            DEFAULT_AUG = torch.nn.Sequential(
                T.RandomHorizontalFlip(),
                T.RandomVerticalFlip(),
                RandomApply(
                    T.GaussianBlur((3, 3), (1.0, 2.0)),
                    p = 0.2
                ),
                #T.Normalize(mean = [0.31955430042667393], std = [0.25145716010016883]),
            )
        if self.args.task == "RIF":
            self.apply_transform =  torch.nn.Sequential(
                T.Normalize(
                    mean=torch.tensor([0.35381392, 0.19792648, 0.109140955]),        
                    std=torch.tensor([0.2853565, 0.1724628, 0.12016233])),
                    ##rif
                    
            )
            DEFAULT_AUG = torch.nn.Sequential(
                RandomApply(
                    T.ColorJitter(0.8, 0.8, 0.8, 0.2),
                    p = 0.3
                ),
                T.RandomGrayscale(p=0.2),
                T.RandomHorizontalFlip(),
                RandomApply(
                    T.GaussianBlur((3, 3), (1.0, 2.0)),
                    p = 0.2
                ),
                #T.RandomResizedCrop((200, 200)),
                T.Normalize(
                    mean=torch.tensor([0.35381392, 0.19792648, 0.109140955]),
                    std=torch.tensor([0.2853565, 0.1724628, 0.12016233])),
            )
        if self.args.task == "EndoPolyp":
            self.apply_transform =  torch.nn.Sequential(
                T.Normalize(
                    mean=torch.tensor([0.4967411, 0.31040248, 0.2248057]),
                    std=torch.tensor([0.3161397, 0.22419964, 0.18269733])),
                    ####End
            )
            DEFAULT_AUG = torch.nn.Sequential(
                RandomApply(
                    T.ColorJitter(0.8, 0.8, 0.8, 0.2),
                    p = 0.3
                ),
                T.RandomGrayscale(p=0.2),
                T.RandomHorizontalFlip(),
                RandomApply(
                    T.GaussianBlur((3, 3), (1.0, 2.0)),
                    p = 0.2
                ),
                #T.RandomResizedCrop((200, 200)),
                T.Normalize(
                    mean=torch.tensor([0.4967411, 0.31040248, 0.2248057]),
                    std=torch.tensor([0.3161397, 0.22419964, 0.18269733])),
            )
        self.augment1 = default(None, DEFAULT_AUG)
        self.augment2 = default(None, self.augment1)
        
    def train_test(self, dataset_train, dataset_test):
        """
        Returns train, validation and test dataloaders for a given dataset
        and user indexes.
        """
        trainloader = DataLoader(DatasetSplit(dataset_train,self.args.task),batch_size=self.args.local_bs, shuffle=True)
        testloader = DataLoader(DatasetSplit(dataset_test,self.args.task),batch_size=1, shuffle=True)
        return trainloader, testloader

    def train_local(self, net, w_glob_keys, global_round,lr=0.00001):
        net.train()
        
        if self.args.optimizer == 'sgd':
            optimizer = torch.optim.SGD(net.parameters(), lr=self.args.lr,
                                        momentum=0.5)
        elif self.args.optimizer == 'adam':
            optimizer = torch.optim.Adam(net.parameters(), lr=self.args.lr,
                                         weight_decay=1e-4)

        local_eps = self.args.local_ep
        epoch_loss = []
        num_updates = 0
        for iter in range(local_eps):
            # for name, param in net.named_parameters():
                # if name in w_glob_keys:
                    # param.requires_grad = True
                # else:
                    # param.requires_grad = False

            batch_loss = []
            for batch_idx, (images, labels) in enumerate(self.trainloader):
                images, labels = images.to(self.args.device), labels.to(self.args.device)
                images = self.apply_transform(images)
                net.zero_grad()
                out = net(images)
                loss = dice_loss(
                                    F.softmax(out, dim=1).float(),
                                    F.one_hot(labels,2).permute(0, 3, 1, 2).float(),
                                    multiclass=True)
                loss = 0.5*loss + 0.5*self.loss_func(out, labels)
                loss.backward()
                optimizer.step()
                num_updates += 1
                batch_loss.append(loss.item())
                # if batch_idx % 2 == 0:
                    # print('\r','| Global Round : {} | Local Epoch : {} | [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                        # global_round, iter, batch_idx * len(images),
                        # len(self.trainloader.dataset),
                        # 100. * batch_idx / len(self.trainloader), loss.item()),end='')
            epoch_loss.append(sum(batch_loss)/len(batch_loss))
        # print("         ")
        return net.state_dict(), sum(epoch_loss) / len(epoch_loss)

    def train_head(self, net, w_glob_keys, global_round,lr=0.00001):
        net.train()
        
        if self.args.optimizer == 'sgd':
            optimizer = torch.optim.SGD(net.parameters(), lr=self.args.lr,
                                        momentum=0.5)
        elif self.args.optimizer == 'adam':
            optimizer = torch.optim.Adam(net.parameters(), lr=self.args.lr,
                                         weight_decay=1e-4)

        local_eps = self.args.local_ep
        epoch_loss = []
        num_updates = 0
        
        for name, param in net.named_parameters():
            if name in w_glob_keys:
                param.requires_grad = False
            else:
                param.requires_grad = True
        for iter in range(local_eps):
            batch_loss = []
            for batch_idx, (images, labels) in enumerate(self.trainloader):
                images, labels = images.to(self.args.device), labels.to(self.args.device)
                images = self.apply_transform(images)
                net.zero_grad()
                out = net(images)
                loss = dice_loss(
                                    F.softmax(out, dim=1).float(),
                                    F.one_hot(labels,2).permute(0, 3, 1, 2).float(),
                                    multiclass=True)
                loss = 0.5*loss + 0.5*self.loss_func(out, labels)
                loss.backward()
                optimizer.step()
                num_updates += 1
                batch_loss.append(loss.item())
                # if batch_idx % 2 == 0:
                    # print('\r','| Global Round : {} | Local Epoch : {} | [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                        # global_round, iter, batch_idx * len(images),
                        # len(self.trainloader.dataset),
                        # 100. * batch_idx / len(self.trainloader), loss.item()),end='')
            epoch_loss.append(sum(batch_loss)/len(batch_loss))
        # print("         ")
        return net.state_dict(), sum(epoch_loss) / len(epoch_loss)


    def train_body(self, net , w_glob_keys, global_round,lr=0.00001):
        net.train()
        
        if self.args.optimizer == 'sgd':
            optimizer = torch.optim.SGD(net.parameters(), lr=self.args.lr,
                                        momentum=0.5)
        elif self.args.optimizer == 'adam': ###self.args.lr
            optimizer = torch.optim.Adam(net.parameters(), lr=0.0001,
                                         weight_decay=1e-4)
        
        local_eps = self.args.local_ep
        epoch_loss = []
        num_updates = 0
        net1 = get_target_encoder(net)
        target_ema_updater = EMA(0.8)##0.8 RIF
        #print(net1)
        for iter in range(local_eps):
            batch_loss = []
            for batch_idx, (images, labels) in enumerate(self.trainloader):
                images, labels = images.to(self.args.device), labels.to(self.args.device)
                augment1,augment2 = self.augment1(images),self.augment2(images)
                net.zero_grad()
                online_out1 = net(augment1)
                online_out2 = net(augment2)
                #print(online_out1.shape,online_out2.shape)
                with torch.no_grad():
                     target_out1 = net1(augment1)
                     target_out2 = net1(augment2)
                     target_out1.detach_()
                     target_out2.detach_()
                loss_one = loss_fn(online_out1, target_out1.detach())
                loss_two = loss_fn(online_out2, target_out2.detach())
                loss = loss_one + loss_two
                loss = loss.mean()
                #print(loss.mean())
                loss.backward()
                optimizer.step()
                #update_moving_average(target_ema_updater,net1,net)
                num_updates += 1
                batch_loss.append(loss.item())
                # if batch_idx % 2 == 0:
                    # print('\r','| Global Round : {} | Local Epoch : {} | [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                        # global_round, iter, batch_idx * len(images),
                        # len(self.trainloader.dataset),
                        # 100. * batch_idx / len(self.trainloader), loss.item()),end='')
            update_moving_average(target_ema_updater,net1,net)
            epoch_loss.append(sum(batch_loss)/len(batch_loss))
        # print("         ")
        return net.state_dict(), sum(epoch_loss) / len(epoch_loss)


    def train_rep(self, net, w_glob_keys_body, w_glob_keys_decoder, global_round,lr=0.00001):
        net.train()
        if self.args.optimizer == 'sgd':
            optimizer = torch.optim.SGD(net.parameters(), lr=self.args.lr,
                                        momentum=0.5)
        elif self.args.optimizer == 'adam':
            optimizer = torch.optim.Adam(net.parameters(), lr=self.args.lr,
                                         weight_decay=1e-4)

        local_eps = self.args.local_ep
        head_eps = local_eps-self.args.local_rep_ep
        epoch_loss = []
        num_updates = 0
        for iter in range(local_eps):

            # for FedRep, first do local epochs for the head
            if iter < head_eps:
                for name, param in net.named_parameters():
                    if name in w_glob_keys_body:
                        param.requires_grad = False
                    else:
                        param.requires_grad = True
            
            # then do local epochs for the representation
            elif iter == head_eps:
                for name, param in net.named_parameters():
                    if name in w_glob_keys_decoder:
                        param.requires_grad = True
                    else:
                        param.requires_grad = False

       
            batch_loss = []
            for batch_idx, (images, labels) in enumerate(self.trainloader):
                images, labels = images.to(self.args.device), labels.to(self.args.device)
                images = self.apply_transform(images)
                net.zero_grad()
                out = net(images)
                loss = dice_loss(
                                    F.softmax(out, dim=1).float(),
                                    F.one_hot(labels,2).permute(0, 3, 1, 2).float(),
                                    multiclass=True)
                loss = 0.5*loss + 0.5*self.loss_func(out, labels)
                loss.backward()
                optimizer.step()
                num_updates += 1
                batch_loss.append(loss.item())
                if batch_idx % 2 == 0:
                    print('\r','| Global Round : {} | Local Epoch : {} | [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                        global_round, iter, batch_idx * len(images),
                        len(self.trainloader.dataset),
                        100. * batch_idx / len(self.trainloader), loss.item()),end='')
            epoch_loss.append(sum(batch_loss)/len(batch_loss))
        print("         ")
        return net.state_dict(), sum(epoch_loss) / len(epoch_loss)

    def inference(self, net, w_locals, w_glob_keys, idx, task, epoch):
        """ Returns the inference accuracy and loss.
        """
        model = copy.deepcopy(net)
        if w_locals is not None:
            w_local = model.state_dict()
            for k in w_locals[idx].keys():
                if w_glob_keys is not None and k not in w_glob_keys:
                    w_local[k] = w_locals[idx][k]
                elif w_glob_keys is None:
                    w_local[k] = w_locals[idx][k]
            model.load_state_dict(w_local)
        model = model.to(self.args.device)
        model.eval()
        loss, total, correct = 0.0, 0.0, 0.0

        with torch.no_grad():
            dice_list = []
            for i,dataset in enumerate(self.testloader):
                
                batchimages,batchLabel = dataset
                batchimages = self.apply_transform(batchimages)
                inputs = batchimages.to(self.args.device)  #######Train
                target = batchLabel.to(self.args.device)
                mask_true = target.long()
                mask_pred = model(inputs)             #得到前向传播的结果
                #print(out.size())
                mask_true = F.one_hot(mask_true, 2).permute(0, 3, 1, 2).float()
                mask_pred = F.one_hot(mask_pred.argmax(dim=1), 2).permute(0, 3, 1, 2).float()
                dice = multiclass_dice_coeff(mask_pred[:, 1:], mask_true[:, 1:], reduce_batch_first=False)
                #print(mask_true.size(),mask_pred.size())
                #print(dice.item())
                dice_list.append(dice.item())
                #print(dice)
                if task == "RIF":
                    if self.idx == 0:
                        print('\r','BinRushed dice:{:.4f}'.format(np.mean(dice_list)),end='')
                    if self.idx == 1:
                        print('\r','Magrabia dice:{:.4f}'.format(np.mean(dice_list)),end='')
                    if self.idx == 2:
                        print('\r','IDRID_Train dice:{:.4f}'.format(np.mean(dice_list)),end='')
                    if self.idx == 3:
                        print('\r','RIM-ONE dice:{:.4f}'.format(np.mean(dice_list)),end='')
                    if self.idx == 4:
                        print('\r','MESSIDOR dice:{:.4f}'.format(np.mean(dice_list)),end='')
                    if self.idx == 5:
                        print('\r','REFUGE_Train dice:{:.4f}'.format(np.mean(dice_list)),end='')
                    if self.idx == 6:
                        print('\r','Drishti_GS1_Train dice:{:.4f}'.format(np.mean(dice_list)),end='')
                if task == "EndoPolyp":
                    if self.idx == 0:
                        print('\r','CVC-ColonDB dice:{:.4f}'.format(np.mean(dice_list)),end='')
                        torch.save(model, "/home/xly/pFred_medical_segmentation/save_model/pFLSC_person/"+str(epoch + 1)+"my_model.pth")
                    if self.idx == 1:
                        print('\r','ETIS-LaribPolypDB dice:{:.4f}'.format(np.mean(dice_list)),end='')
                    if self.idx == 2:
                        print('\r','CVC-ClinicDB dice:{:.4f}'.format(np.mean(dice_list)),end='')
                    if self.idx == 3:
                        print('\r','Kvasir-SEG dice:{:.4f}'.format(np.mean(dice_list)),end='')
                if task == "PMR":
                    if self.idx == 0:
                        print('\r','BIDMC dice:{:.4f}'.format(np.mean(dice_list)),end='')
                    if self.idx == 1:
                        print('\r','BMC dice:{:.4f}'.format(np.mean(dice_list)),end='')
                    if self.idx == 2:
                        print('\r','HK dice:{:.4f}'.format(np.mean(dice_list)),end='')
                    if self.idx == 3:
                        print('\r','I2CVB dice:{:.4f}'.format(np.mean(dice_list)),end='')
                    if self.idx == 4:
                        print('\r','RUNMC dice:{:.4f}'.format(np.mean(dice_list)),end='')
                    if self.idx == 5:
                        print('\r','UCL dice:{:.4f}'.format(np.mean(dice_list)),end='')
            print("         ")


    def inference_average(self, net, w_locals, w_glob_keys, idx):
        """ Returns the inference accuracy and loss.
        """
        model = copy.deepcopy(net)
        if w_locals is not None:
            w_local = model.state_dict()
            for k in w_locals[idx].keys():
                if w_glob_keys is not None and k not in w_glob_keys:
                    w_local[k] = w_locals[idx][k]
                elif w_glob_keys is None:
                    w_local[k] = w_locals[idx][k]
            model.load_state_dict(w_local)
        model = model.to(self.args.device)
        model.eval()
        loss, total, correct = 0.0, 0.0, 0.0
        all_loss = []
        with torch.no_grad():
            dice_list = []
            for i,dataset in enumerate(self.testloader):
                
                batchimages,batchLabel = dataset
                batchimages = self.apply_transform(batchimages)
                inputs = batchimages.to(self.args.device)  #######Train
                target = batchLabel.to(self.args.device)
                mask_true = target.long()
                mask_pred = model(inputs)             #得到前向传播的结果
                loss = dice_loss(
                                    F.softmax(mask_pred, dim=1).float(),
                                    F.one_hot(mask_true,2).permute(0, 3, 1, 2).float(),
                                    multiclass=True)
                loss = 0.5*loss + 0.5*self.loss_func(mask_pred, mask_true)
                
                
                #print(out.size())
                mask_true = F.one_hot(mask_true, 2).permute(0, 3, 1, 2).float()
                mask_pred = F.one_hot(mask_pred.argmax(dim=1), 2).permute(0, 3, 1, 2).float()
                dice = multiclass_dice_coeff(mask_pred[:, 1:], mask_true[:, 1:], reduce_batch_first=False)
                #print(mask_true.size(),mask_pred.size())
                #print( dice.item())
                dice_list.append(dice.item())
                all_loss.append(loss.item())
                #print(dice)
            return np.mean(dice_list),np.mean(all_loss)